import os
import os.path
import shutil
import sys

if sys.version_info.major > 2:  # pragma: no cover
    from urllib.parse import urlparse
else:
    from urlparse import urlparse

from pwman.data import factory
from pwman.util import config
from pwman.util.callback import Callback


def get_ui_platform(platform):  # pragma: no cover
    if 'darwin' in platform:
        from pwman.ui.mac import PwmanCliMac as PwmanCli
        OSX = True
    elif 'win' in platform:
        from pwman.ui.win import PwmanCliWin as PwmanCli
        OSX = False
    else:
        from pwman.ui.cli import PwmanCli
        OSX = False

    return PwmanCli, OSX

PwmanCliNew, OSX = get_ui_platform(sys.platform)


class DummyCallback(Callback):

    def getinput(self, question):
        return b'12345'

    def getsecret(self, question):
        return b'12345'


class DummyCallback2(Callback):

    def getinput(self, question):
        return b'newsecret'

    def getsecret(self, question):
        return b'wrong'


class DummyCallback3(Callback):

    def getinput(self, question):
        return b'newsecret'

    def getsecret(self, question):
        ans = b'12345'
        return ans


class DummyCallback4(Callback):

    def getinput(self, question):
        return b'newsecret'

    def getsecret(self, question):
        return b'newsecret'

db =  ".".join(("pwman","test", sys.version.split(" " ,1)[0], "db"))
testdb = os.path.abspath(os.path.join(os.path.dirname(__file__), db))

config.default_config['Database'] = {'type': 'sqlite',
                                     'filename': testdb,
                                     'dburi': os.path.join(
                                         'sqlite:///', testdb)
                                     }

dc = """
[Global]
xsel = /usr/bin/xsel
colors = yes
umask = 0100
cls_timeout = 5

[Database]
type = SQLite
"""


class SetupTester(object):

    def __init__(self, dbver=None, filename=None, dburi=None):

        f = open(os.path.join(os.path.dirname(__file__), 'test.conf'), 'w')
        f.write(dc)
        f.close()
        self.configp = config.Config(os.path.join(os.path.dirname(__file__),
                                                  "test.conf"),
                                     config.default_config)

        self.configp.set_value('Database', 'dburi',
                               'sqlite://' + testdb
                               )

        if not OSX:
            self.xselpath = shutil.which("xsel") or ""
            self.configp.set_value("Global", "xsel", self.xselpath)
        else:
            self.xselpath = "xsel"

        self.dbver = dbver
        self.dburi = self.configp.get_value('Database', 'dburi')

    def clean(self):
        dbfile = self.configp.get_value('Database', 'filename')
        dburi = urlparse(self.configp.get_value('Database', 'dburi')).path
        try:
            if os.path.exists(dbfile):
                os.remove(dbfile)

            if os.path.exists(dburi):
                os.remove(dburi)
        except PermissionError:
            pass

        if os.path.exists(os.path.join(os.path.dirname(__file__),
                                       'testing_config')):
            os.remove(os.path.join(os.path.dirname(__file__),
                                   'testing_config'))

        if os.path.exists(os.path.join(os.path.dirname(__file__),
                                       'test.conf')):
            os.remove(os.path.join(os.path.dirname(__file__),
                                   'test.conf'))

    def create(self):
        db = factory.createdb(self.dburi, self.dbver)
        self.cli = PwmanCliNew(db, self.xselpath, DummyCallback,
                               config_parser=self.configp)



# ============================================================================
# This file is part of Pwman3.
#
# Pwman3 is free software; you can redistribute iut and/or modify
# it under the terms of the GNU General Public License, version 2
# as published by the Free Software Foundation;
#
# Pwman3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pwman3; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ============================================================================
# Copyright (C) 2015-2017 Oz Nahum Tiram <nahumoz@gmail.com>
# ============================================================================
import os
import unittest
from .test_crypto_engine import give_key, DummyCallback
from urllib.parse import urlparse

import psycopg2 as pg
from pwman.data.drivers.postgresql import PostgresqlDatabase
from pwman.util.crypto_engine import CryptoEngine
##
# testing on linux host
# su - postgres
# psql
# postgres=# CREATE USER tester WITH PASSWORD '123456';
# postgres=# grant ALL ON DATABASE pwman to tester;
#
##

POSTGRESHOST = os.getenv("POSTGRESHOST", "localhost")
DBURI = "postgresql://tester:123456@%s/pwman" % POSTGRESHOST


class TestPostGresql(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        u = DBURI
        u = urlparse(u)
        # password required, for all hosts
        # u = "postgresql://<user>:<pass>@localhost/pwman"
        self.db = PostgresqlDatabase(u)
        self.db._open()

    @classmethod
    def tearDownClass(self):
        self.db._cur.execute("DROP TABLE LOOKUP")
        self.db._cur.execute("DROP TABLE TAG")
        self.db._cur.execute("DROP TABLE NODE")
        self.db._cur.execute("DROP TABLE DBVERSION")
        self.db._cur.execute("DROP TABLE CRYPTO")
        self.db._con.commit()

    def test_1_con(self):
        self.assertIsInstance(self.db._cur, pg._psycopg.cursor)

    def test_2_create_tables(self):
        self.db._create_tables()
        # invoking this method a second time should not raise an exception
        self.db._create_tables()

    def test_3_load_key(self):
        self.db.savekey('SECRET$6$KEY')
        secretkey = self.db.loadkey()
        self.assertEqual(secretkey, b'SECRET$6$KEY')

    def test_4_save_crypto(self):
        self.db.save_crypto_info(b"TOP", b"SECRET")
        secretkey = self.db.loadkey()
        self.assertEqual(secretkey, b'TOP$6$SECRET')
        row = self.db.fetch_crypto_info()
        self.assertEqual(list(map(bytearray, row)),
                         [bytearray(b'TOP'), bytearray(b'SECRET')])

    def test_5_add_node(self):
        innode = [b"TBONE", b"S3K43T", b"example.org", b"some note",
                  [b"footag", b"bartag"]]
        self.db.add_node(innode)
        outnode = self.db.getnodes([1])[0]
        self.assertEqual(innode[:-1] + [t for t in innode[-1]], outnode[1:])

    def test_6_list_nodes(self):
        ret1 = self.db.lazy_list_node_ids()
        self.assertEqual(list(ret1), [1])
        ret2 = self.db.lazy_list_node_ids(b"footag")
        self.assertEqual(list(ret2), [1])

    def test_6a_list_tags(self):
        ret = self.db.listtags()
        self.assertListEqual(ret, [b'footag', b'bartag'])

    def test_6b_get_nodes(self):
        ret = self.db.getnodes([1])
        retb = self.db.getnodes([])
        self.assertListEqual(ret, retb)

    def test_7_get_or_create_tag(self):
        s = self.db._get_or_create_tag(b"SECRET")
        s1 = self.db._get_or_create_tag(b"SECRET")

        self.assertEqual(s, s1)

    def test_7a_clean_orphans(self):

        self.db._clean_orphans()
        rv = self.db._get_tag("SECRET")
        self.assertIsNone(rv)

    def test_8_remove_node(self):
        self.db.removenodes([1])
        nids = list(self.db.lazy_list_node_ids())
        self.assertEqual(len(nids), 0)

    def test_9_check_db_version(self):

        dburi = DBURI
        v = self.db.check_db_version(dburi)
        self.assertEqual(str(v), '0.6')
        self.db._cur.execute("DROP TABLE DBVERSION")
        self.db._con.commit()
        v = self.db.check_db_version(dburi)
        self.assertEqual(str(v), '0.6')
        self.db._cur.execute("CREATE TABLE DBVERSION("
                             "VERSION TEXT NOT NULL DEFAULT {}"
                             ")".format('0.6'))
        self.db._con.commit()


if __name__ == '__main__':

    ce = CryptoEngine.get()
    ce.callback = DummyCallback()
    ce.changepassword(reader=give_key)
    unittest.main(verbosity=2, failfast=True)



# ============================================================================
# This file is part of Pwman3.
#
# Pwman3 is free software; you can redistribute iut and/or modify
# it under the terms of the GNU General Public License, version 2
# as published by the Free Software Foundation;
#
# Pwman3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pwman3; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ============================================================================
# Copyright (C) 2012-2017 Oz Nahum Tiram <nahumoz@gmail.com>
# ============================================================================
import os
import unittest
from pwman.data.drivers.sqlite import SQLite
from pwman.data.nodes import Node
from pwman.util.crypto_engine import CryptoEngine
from .test_crypto_engine import give_key, DummyCallback


class TestSQLite(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        cls.db.close()
        for item in ('test.db',):
            try:
                os.remove(item)
            except OSError:
                continue

    @classmethod
    def setUp(cls):
        cls.db = SQLite('test.db')
        cls.db._open()

    def test_1_create_tables(self):
        self.db._create_tables()
        self.db._con.commit()
        # the method _open calls _create_tables
        self.db.save_crypto_info("foo", "bar")
        self.db._create_tables()

    def test_1a_create_tables(self):
        self.db._create_tables()

    def test_2_crypto_info(self):
        self.db._create_tables()
        self.db.save_crypto_info("foo", "bar")
        f = self.db.fetch_crypto_info()
        self.assertListEqual([u'foo', u'bar'], list(f))

    def test_3_add_node(self):
        node = Node(clear_text=True,
                    **{'username': "alice", 'password': "secret",
                       'url': "wonderland.com",
                       'notes': "a really great place",
                       'tags': ['foo', 'bar']})
        self.db.add_node(node)
        rv = self.db._cur.execute("select * from node")
        ce = CryptoEngine.get()
        res = rv.fetchone()
        self.assertEqual(ce.decrypt(res[1]), b"alice")

    def test_4_test_tags(self):
        node = Node(clear_text=True,
                    **{'username': u"alice", 'password': u"secret",
                       'url': u"wonderland.com",
                       'notes': u"a really great place",
                       'tags': ['foo', 'bar']})
        ce = CryptoEngine.get()
        self.db._get_or_create_tag(node._tags[0])
        self.assertEqual(1, self.db._get_or_create_tag(node._tags[0]))
        rv = self.db._get_or_create_tag(ce.encrypt(b'baz'))
        self.assertEqual(3, rv)
        self.db._con.commit()

    def test_5_test_lookup(self):
        self.db._cur.execute('SELECT nodeid, tagid FROM LOOKUP')
        rows = self.db._cur.fetchall()
        self.assertEqual(2, len(rows))

    def test_6_listnodes(self):
        node = Node(clear_text=True,
                    **{'username': u"hatman", 'password': u"secret",
                       'url': u"wonderland.com",
                       'notes': u"a really great place",
                       'tags': [u'baz', u'bar']})
        self.db.add_node(node)
        ids = self.db.lazy_list_node_ids()
        self.assertEqual(2, len(list(ids)))

    def test_7_listnodes_w_filter(self):
        ce = CryptoEngine.get()
        # the tag 'bar' is found in a node created in:
        # test_3_add_node
        # test_6_listnodes

        tag = ce.encrypt(b'bar')
        rv = self.db.lazy_list_node_ids(filter=tag)
        self.assertEqual(len(list(rv)), 2)
        tag = ce.encrypt(b'baz')
        # the tag 'baz' is found in a node created in
        # test_6_listnodes
        rv = self.db.lazy_list_node_ids(filter=tag)
        self.assertEqual(len(list(rv)), 1)

    def test_8_getnodes(self):
        nodes = self.db.getnodes([1, 2])
        self.assertEqual(len(list(nodes)), 2)

    def test_9_editnode(self):
        # delibertly insert clear text into the database
        ce = CryptoEngine.get()
        tags = [ce.encrypt("foo"), ce.encrypt("auto")]
        node = {'user': 'transparent', 'password': 'notsecret',
                'tags': tags}
        self.db.editnode('2', **node)
        self.db._cur.execute('SELECT USER, PASSWORD FROM NODE WHERE ID=2')
        rv = self.db._cur.fetchone()
        self.assertEqual(rv, ('transparent', 'notsecret'))
        node = {'user': 'modify', 'password': 'notsecret',
                'tags': tags}
        # now the tags bank and baz are orphan ...
        # what happens? it should be completely removed.
        # To spare IO we only delete orphand tags when
        # db.close is called.
        self.db.editnode('2', **node)

    def test_9_test_no_orphans(self):
        self.db._clean_orphans()
        self.db._con.commit()
        ce = CryptoEngine.get()
        tags = None
        while not tags:
            tags = self.db._cur.execute('SELECT * FROM tag').fetchall()
        tags_clear = [ce.decrypt(tag[1]) for tag in tags]
        self.assertNotIn(b"baz", tags_clear)

    def test_a10_test_listtags(self):
        """there should be only 3 tags left"""
        tags = self.db.listtags()
        self.assertEqual(3, len(list(tags)))

    def test_a11_test_rmnodes(self):
        for n in [1, 2]:
            self.db.removenodes([n])
        rv = self.db._cur.execute("select * from node").fetchall()
        self.assertListEqual(rv, [])

    def test_a12_test_savekey(self):
        ce = CryptoEngine.get()
        self.db.savekey(ce.get_salt_digest())
        self.assertEqual(ce.get_salt_digest(), self.db.loadkey())


if __name__ == '__main__':

    ce = CryptoEngine.get()
    ce.callback = DummyCallback()
    ce.changepassword(reader=give_key)
    unittest.main(verbosity=2, failfast=True)



import unittest

from .test_postgresql import TestPostGresql
from .test_mongodb import TestMongoDB
from .test_mysql import TestMySQLDatabase
from .test_pwman import suite


def complete_suite(suite):

    loader = unittest.TestLoader()
    suite.addTest(loader.loadTestsFromTestCase(TestPostGresql))
    suite.addTest(loader.loadTestsFromTestCase(TestMySQLDatabase))
    suite.addTest(loader.loadTestsFromTestCase(TestMongoDB))

    return suite


if __name__ == '__main__':
    test_suite = complete_suite(suite())
    unittest.TextTestRunner(verbosity=2, failfast=True).run(test_suite)



# ============================================================================
# This file is part of Pwman3.
#
# Pwman3 is free software; you can redistribute iut and/or modify
# it under the terms of the GNU General Public License, version 2
# as published by the Free Software Foundation;
#
# Pwman3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pwman3; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ============================================================================
# Copyright (C) 2015 Oz Nahum Tiram <nahumoz@gmail.com>
# ============================================================================

"""
setting up mysql for testing

$ mysql -u root -h localhost -p

mysql > CREATE DATABASE pwmantest;
mysql > CREATE USER pwman IDENTIFIED BY '123456';
mysql > GRANT ALL on pwmantest.* to 'pwman'@'localhost';
"""
import os
import unittest
import time

from urllib.parse import urlparse

import pymysql

from pwman.data.drivers.mysql import MySQLDatabase
from pwman.util.crypto_engine import CryptoEngine

from .test_crypto_engine import give_key, DummyCallback


MYSQLHOST = os.getenv("MYSQLHOST", "localhost")
DBURI = "mysql://pwman:123456@%s:3306/pwmantest" % MYSQLHOST


class TestMySQLDatabase(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        u = DBURI
        u = urlparse(u)
        # password required, for all hosts
        # u = "mysql://<user>:<pass>@localhost/pwmantest"
        for i in range(60):
            try:
                self.db = MySQLDatabase(u)
                self.db._open()
            except:
                time.sleep(1)

    @classmethod
    def tearDownClass(self):
        for table in ['LOOKUP', 'TAG', 'NODE', 'DBVERSION', 'CRYPTO']:
            try:
                self.db._cur.execute("DROP TABLE {}".format(table))
            except Exception:
                pass
            self.db._con.commit()

    def test_1_con(self):
        self.assertIsInstance(self.db._con, pymysql.connections.Connection)

    def test_2_create_tables(self):
        self.db._create_tables()
        # invoking this method a second time should not raise an exception
        self.db._create_tables()

    def test_3_load_key(self):
        self.db.savekey('SECRET$6$KEY')
        secretkey = self.db.loadkey()
        self.assertEqual(secretkey, 'SECRET$6$KEY')

    def test_4_save_crypto(self):
        self.db.save_crypto_info("TOP", "SECRET")
        secretkey = self.db.loadkey()
        self.assertEqual(secretkey, 'TOP$6$SECRET')
        row = self.db.fetch_crypto_info()
        self.assertEqual(row, ('TOP', 'SECRET'))

    def test_5_add_node(self):
        innode = ["TBONE", "S3K43T", "example.org", "some note",
                  ["bartag", "footag"]]
        self.db.add_node(innode)

        outnode = list(self.db.getnodes([1]))[0]
        self.assertEqual(innode[:-1] + [t for t in innode[-1]], outnode[1:])

    def test_6_list_nodes(self):
        ret = self.db.lazy_list_node_ids()
        self.assertEqual(list(ret), [1])
        ret = self.db.lazy_list_node_ids("footag")
        self.assertEqual(list(ret), [1])

    def test_6a_list_tags(self):
        ret = self.db.listtags()
        self.assertListEqual(ret, ['bartag', 'footag'])

    def test_6b_get_nodes(self):
        ret = list(self.db.getnodes([1]))
        self.assertListEqual(ret[0], [1, "TBONE", "S3K43T", "example.org",
                                      "some note",
                                      "bartag", "footag"])

    def test_7_get_or_create_tag(self):
        s = self.db._get_or_create_tag("SECRET")
        s1 = self.db._get_or_create_tag("SECRET")

        self.assertEqual(s, s1)

    def test_7a_clean_orphans(self):

        self.db._clean_orphans()
        rv = self.db._get_tag("SECRET")
        self.assertIsNone(rv)

    def test_8_remove_node(self):
        self.db.removenodes([1])
        n = list(self.db.lazy_list_node_ids())
        self.assertEqual(len(n), 0)

    def test_9_check_db_version(self):

        dburi = DBURI
        v = self.db.check_db_version(urlparse(dburi))
        self.assertEqual(v, '0.6')
        self.db._cur.execute("DROP TABLE DBVERSION")
        self.db._con.commit()
        v = self.db.check_db_version(urlparse(dburi))
        self.assertEqual(v, '0.6')
        self.db._cur.execute("CREATE TABLE DBVERSION("
                             "VERSION TEXT NOT NULL) ")
        self.db._con.commit()


if __name__ == '__main__':

    ce = CryptoEngine.get()
    ce.callback = DummyCallback()
    ce.changepassword(reader=give_key)
    unittest.main(verbosity=2, failfast=True)



# ============================================================================
# This file is part of Pwman3.
#
# Pwman3 is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2
# as published by the Free Software Foundation;
#
# Pwman3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pwman3; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ============================================================================
# Copyright (C) 2014 Oz Nahum Tiram <nahumoz@gmail.com>
# ============================================================================
import os
import unittest
import sys
from io import StringIO, BytesIO

from pwman.util.crypto_engine import CryptoEngine
from .test_crypto_engine import give_key, DummyCallback
from pwman.data.database import __DB_FORMAT__
from .test_tools import (SetupTester, testdb)
from pwman.data import factory
from pwman.data.nodes import Node


class dummy_stdin(object):

    def __init__(self):
        self.idx = -1
        self.ans = ['4', 'some fucking notes', 'X']

    def __call__(self, msg):
        self.idx += 1
        return self.ans[self.idx]


class TestBaseUI(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for item in (testdb, 'foo.csv', 'pwman-export.csv'):
            try:
                os.unlink(item)
            except OSError:
                continue

        SetupTester().clean()

    def setUp(self):
        "test that the right db instance was created"
        dbver = __DB_FORMAT__
        self.db = factory.createdb('sqlite://' + testdb, dbver)
        self.tester = SetupTester(dbver, testdb)
        self.tester.create()

    def tearDown(self):
        pass

    def test_get_tags(self):
        sys.stdin = StringIO("foo,bar,baz")
        tags = self.tester.cli._get_tags(reader=lambda: "foo,bar,baz")
        self.assertListEqual(['foo', 'bar', 'baz'], tags)
        sys.stdin = sys.__stdin__

    def test_1_do_new(self):
        sys.stdin = BytesIO((b"alice\nsecret\nexample.com\nsome notes"
                             b"\nfoo,bar,baz"))
        _node = self.tester.cli._do_new('')

        sys.stdin = sys.__stdin__
        self.assertListEqual([b'foo', b'bar', b'baz'], [t for t
                                                        in _node.tags])
        node_ids = self.tester.cli._db.lazy_list_node_ids()
        self.assertListEqual([1], list(node_ids))
        nodes = list(self.tester.cli._db.getnodes([1]))
        ce = CryptoEngine.get()
        user = ce.decrypt(nodes[0][1])
        self.assertTrue(user, 'alice')
        tags = nodes[-1]
        for idx, t in enumerate(['foo', 'bar', 'baz']):
            self.assertTrue(t, tags[idx])

    def test_2_do_list(self):
        self.output = StringIO()
        sys.stdout = self.output
        self.tester.cli.do_list('')
        self.tester.cli.do_list('foo')
        self.tester.cli.do_list('bar')
        sys.stdout = sys.__stdout__
        self.output.getvalue()

    def test_3_do_export(self):
        self.tester.cli.do_export("{'filename':'foo.csv'}")
        with open('foo.csv') as f:
            lines = f.readlines()
        # on windows there is an extra empty line in the exported file
        self.assertIn('alice;example.com;secret;some notes;foo,bar,baz\n',
                      lines)

    def test_3a_do_export(self):
        self.tester.cli.do_export("f")
        with open('pwman-export.csv') as f:
            lines = f.readlines()

        self.assertIn('alice;example.com;secret;some notes;foo,bar,baz\n',
                      lines)

    def test_4_do_forget(self):
        self.tester.cli.do_forget('')
        ce = CryptoEngine.get()
        self.assertIsNone(ce._cipher)

    def test_5_do_print(self):
        v = StringIO()
        sys.stdout = v
        self.tester.cli.do_print('1')
        self.assertIn('\x1b[31mUsername:\x1b[0m alice', v.getvalue())
        self.tester.cli.do_print('a')
        self.assertIn("print accepts only a single ID ...", v.getvalue())

        sys.stdout = sys.__stdout__

    def test_6_do_tags(self):
        v = StringIO()
        sys.stdout = v
        self.tester.cli.do_tags('1')
        v = v.getvalue()
        for t in ['foo', 'bar', 'baz']:
            t in v
        sys.stdout = sys.__stdout__

    def test_7_get_ids(self):
        # used by do_cp or do_open,
        # this spits many time could not understand your input
        self.assertEqual([1], self.tester.cli._get_ids('1'))
        self.assertListEqual([1, 2, 3, 4, 5], self.tester.cli._get_ids('1-5'))
        self.assertListEqual([], self.tester.cli._get_ids('5-1'))
        self.assertListEqual([], self.tester.cli._get_ids('5x-1'))
        self.assertListEqual([], self.tester.cli._get_ids('5x'))
        self.assertListEqual([], self.tester.cli._get_ids('5\\'))

    def test_8_do_edit_1(self):
        node = list(self.tester.cli._db.getnodes([1]))[0]
        node = node[1:5] + [node[5:]]
        node = Node.from_encrypted_entries(*node)
        sys.stdin = StringIO(("1\nfoo\nx\n"))
        self.tester.cli.do_edit('1')
        v = StringIO()
        sys.stdin = sys.__stdin__
        sys.stdout = v
        self.tester.cli.do_print('1')
        self.assertIn('\x1b[31mUsername:\x1b[0m foo', v.getvalue())

    def test_8_do_edit_2(self):
        node = list(self.tester.cli._db.getnodes([1]))[0]
        node = node[1:5] + [node[5:]]
        node = Node.from_encrypted_entries(*node)
        sys.stdin = StringIO(("2\ns3kr3t\nx\n"))
        self.tester.cli.do_edit('1')
        v = StringIO()
        sys.stdin = sys.__stdin__
        sys.stdout = v
        self.tester.cli.do_print('1')
        self.assertIn('\x1b[31mPassword:\x1b[0m s3kr3t', v.getvalue())

    def test_9_do_delete(self):
        self.assertIsNone(self.tester.cli._do_rm('x'))
        sys.stdin = StringIO("y\n")
        self.tester.cli.do_rm('1')
        sys.stdin = sys.__stdin__
        sys.stdout = StringIO()
        self.tester.cli.do_ls('')
        self.assertNotIn('alice', sys.stdout.getvalue())
        sys.stdout = sys.__stdout__

    def test_10_do_info(self):
        self.output = StringIO()
        sys.stdout = self.output
        self.tester.cli.do_info(b'')
        self.assertIn(testdb, sys.stdout.getvalue())


if __name__ == '__main__':

    ce = CryptoEngine.get()
    ce.callback = DummyCallback()
    ce.changepassword(reader=give_key)
    unittest.main(verbosity=3, failfast=True)



# ============================================================================
# This file is part of Pwman3.
#
# Pwman3 is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2
# as published by the Free Software Foundation;
#
# Pwman3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pwman3; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ============================================================================
# Copyright (C) 2018 Oz N Tiram <nahumoz@gmail.com>
# ============================================================================
# Copyright (C) 2006 Ivan Kelly <ivan@ivankelly.net>
# ============================================================================
import os
import platform
import sys
from functools import lru_cache


from configparser import (ConfigParser, ParsingError, NoOptionError,
                          NoSectionError, MissingSectionHeaderError)

# XDG code taken from xdg.py
# https://github.com/srstevenson/xdg/blob/master/xdg.py
# Copyright © 2016-2018 Scott Stevenson <scott@stevenson.io>


def _getenv(variable: str, default: str) -> str:
    """Get an environment variable.
    Parameters
    ----------
    variable : str
        The environment variable.
    default : str
        A default value that will be returned if the environment
        variable is unset or empty.
    Returns
    -------
    str
        The value of the environment variable, or the default value.
    """
    return os.environ.get(variable) or default


XDG_CACHE_HOME = _getenv('XDG_CACHE_HOME',
                         os.path.expandvars(os.path.join('$HOME', '.cache')))
XDG_CONFIG_DIRS = _getenv('XDG_CONFIG_DIRS', '/etc/xdg').split(':')
XDG_CONFIG_HOME = _getenv('XDG_CONFIG_HOME',
                          os.path.expandvars(os.path.join('$HOME', '.config')))
XDG_DATA_DIRS = _getenv('XDG_DATA_DIRS',
                        '/usr/local/share/:/usr/share/').split(':')
XDG_DATA_HOME = _getenv('XDG_DATA_HOME',
                        os.path.expandvars(
                            os.path.join('$HOME', '.local', 'share')))
XDG_RUNTIME_DIR = os.getenv('XDG_RUNTIME_DIR')


@lru_cache(maxsize=None)
def find_config_dir(appname):
    """
    Backward compatibly config dir finder

    If ~/.appname is not found define a new XDG compat one
    """
    config_dir = os.path.expanduser("~/.%s" % appname)

    if os.path.exists(config_dir):
        return config_dir, config_dir
    elif platform.system() == 'Windows':
        app_data = os.path.expandvars(os.path.join('$APPDATA', appname))
        return app_data, app_data
    else:
        return (os.path.join(XDG_CONFIG_HOME, appname),
                os.path.join(XDG_DATA_HOME, appname))


config_dir, data_dir = find_config_dir('pwman')


default_config = {
    'Global':
    {'umask': ('0100 # The umask in which database and '
               'configuration files are written.'),
     'colors': ('yes # set to `no` to supress color in output. '
                'This is useful for breil terminals'),
     'cls_timeout': ('5 # Number of seconds before the screen is'
                     ' clean after a print. '
                     'Set to 0 to disable.'),
     'cp_timeout': ('10 # number of seconds before the clipboard'
                    ' is erased.'),
     'save': ('True #  whether the Configuring file should be '
              'saved'),
     'lock_timeout': '600'
     },
    'Database': {'dburi': ('sqlite://$XDG_DATA_HOME/pwman.db # ')},
    'Readline': {'history': os.path.join(data_dir,
                                         'history')},
    'Crypto': {'supress_warning': 'no'},
    'Updater': {'supress_version_check': ('no # set to `yes` to supress check '
                                          'for newer versions of pwman3'),
                'client_info': ('"" #  sha256 digest of host name and '
                                'username, used for identifying the client'),
                },
    'UI': {"URL_Length":
           '22 # the max length of URL to show. Longer URLs are trimmed',
           "URL_pad":
               '25 # the padding of the URL in `line_format`',
           "user_pad":
               '25 # the padding of the user name in `line_format`',
           "tag_pad":
               '20 # the padding of tags in the `line_format',
           "line_format":
               "{ID:<3} {USER:<{user_pad}}{URL:<{url_pad}}{Tags:<{tag_pad}}",
           }
    }

if 'win' in sys.platform:
    default_config['Database']['dburi'] = default_config['Database']['dburi'].replace("\\", "/")  # noqa


class ConfigException(Exception):
    """Basic exception for config."""
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "{}: {}".format(self.__class__.__name__,
                               self.message)  # pragma: no cover


class ConfigNoConfigException(ConfigException):
    pass


class Config:

    """
    The configuration of pwman is done with an `ini` file found in
    XDG_CONFIG_HOME on Unix systems.
    On windows the configuration is found in ``%APPDATA/pwman/%``
    The following describe the possible sections in the file and the default
    values of each parameter:

    =====================    ===========
    **Section**              *Readline*
    ---------------------    -----------
    history                  path to the file containing history of commands
                             typed
    ---------------------    -----------
    **Section**              *Global*
    ---------------------    -----------
    save                     True or False - whether the Configuring file
                             should be saved
    ---------------------    -----------
    colors                   yes or no - If set to *no*, no colors used in
                             output. This is useful for breil terminals.
    ---------------------    -----------
    cp_timeout               Number of seconds before the clipboard is erased.
    ---------------------    -----------
    cls_timeout              Number of seconds before the screen is clean after
                             a print. Set to 0 to disable.
    ---------------------    -----------
    umask                    The umask in which database and configuration
                             files are written.
    ---------------------    -----------
    xsel                     path to the xsel binary
                             (Linux\\BSD only).
    ---------------------    -----------
    lock_timeout             sets the period (in secods) after which the
                             database is locked.
    ---------------------    -----------
    **Section**              *Database*
    ---------------------    -----------
    dburi                    Database URI conforming to `RFC3986`.
                             SQLite, Postgreql, MySQL and MongoDB are currently
                             supported.

                             SQLite example:
                                 `sqlite:///path/to/your/db`

                             Postgreql example:
                                 `postgresql://<user>:<pass>@<host[:port]>/<database>`

                             MySQL example:
                                 `mysql://<user>:<pass>@<host[:port]>/<database>`

                             MongoDB example:
                                 `mongodb://<user>:<pass>@<host[:port]>/<database>`
    ---------------------    -----------
    **Section**              *Updater*
    ---------------------    -----------
    supress_version_check    yes or no - check for newer versions of pwman3
    ---------------------    -----------
    client_info              sha256 digest of host name and username,
                             used for identifying the client
    ---------------------    -----------
    **Section**              *UI*
    ---------------------    -----------
    URL_length               22  - the max length of URL to show.
                                   Longer URLs are trimmed
    ---------------------    -----------
    URL_pad                  25  - the padding of the URL in `line_format`
    ---------------------    -----------
    user_pad                 25  - the padding of the user name in
                                   `line_format`
    ---------------------    -----------
    tag_pad                  20  - the padding of tags in the `line_format`
    ---------------------    -----------
    line_format              `{ID:<3} {USER:<{user_pad}}{URL:<{url_pad}}{Tags:<{tag_pad}}`  # noqa
    =====================    ===========
    """

    def __init__(self, filename=None, defaults=None, **kwargs):

        self.filename = filename
        self.parser = self._load(defaults)

    def _load(self, defaults):
        defaults = defaults or default_config
        parser = ConfigParser()
        try:
            with open(self.filename) as f:
                try:
                    parser.read_file(f)
                except (ParsingError, MissingSectionHeaderError) as e:
                    raise ConfigException(e)
        except IOError:
            self._self_write_new_conf(self.filename, defaults, parser)

        self._add_defaults(defaults, parser)
        return parser

    def _self_write_new_conf(self, filename, defaults, parser):
        self.parser = parser
        self._add_defaults(defaults, parser)
        self.save()

    def _add_defaults(self, defaults, parser):
        loc_defaults = defaults.copy()
        loc_defaults.update({'Database':
                             {
                              'dburi':  'sqlite://' + os.path.join(data_dir,
                                                                   'pwman.db')},  # noqa
                              'Readline': {'history': os.path.join(data_dir,
                                                                   'history')}})  # noqa
        for section, options in loc_defaults.items():
            if not parser.has_section(section):
                parser.add_section(section)

            for key, value in options.items():
                if not parser.has_option(section, key):
                    parser.set(section, key, value)

    def get_value(self, section, name):
        try:
            return self.parser.get(section, name)
        except (NoOptionError, NoSectionError):  # pragma: no cover
            return ''

    def set_value(self, section, name, value):
        self.parser.set(section, name, value)

    def save(self):
        if "False" not in self.get_value("Global", "Save"):
            with open(self.filename, "w") as fp:
                self.parser.write(fp)


def get_pass_conf(config):
    ascii_lowercase = config.get_value("Generator",
                                       "ascii_lowercase").lower() == 'true'
    ascii_uppercase = config.get_value("Generator",
                                       "ascii_uppercase").lower() == 'true'
    ascii_digits = config.get_value("Generator",
                                    "ascii_digits").lower() == 'true'
    ascii_punctuation = config.get_value("Generator",
                                         "ascii_punctuation").lower() == 'true'
    return ascii_lowercase, ascii_uppercase, ascii_digits, ascii_punctuation
